import base64

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding


def rsa_encrypt(public_key_pem, data):
    """
    使用给定的RSA公钥加密数据
    :param public_key_pem: RSA公钥的PEM格式字符串
    :param data: 要加密的原始数据（字节类型）
    :return: 加密后的数据（字节类型，Base64编码后的结果方便展示和存储）
    """
    try:
        # 加载公钥
        public_key = serialization.load_pem_public_key(
            public_key_pem.encode(), backend=default_backend()
        )
        # 使用OAEP填充方式进行加密（推荐用于RSA加密的填充方式）
        ciphertext = public_key.encrypt(
            data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        # 将加密后的字节数据进行Base64编码，方便后续存储、传输等处理
        return base64.b64encode(ciphertext)
    except Exception as e:
        print("rsa_encrypt error: {}".format(e))
        raise



import json
import os

import click
from prettytable import PrettyTable

from .client import Client

cred_path = os.path.join(os.path.expanduser("~"), ".tikit/credential")


def pretty_print_dict_list(l):
    if not l:
        return None
    table = PrettyTable()
    cols = l[0].keys()
    for col in cols:
        table.add_column(col, [])
    for dct in l:
        table.add_row([dct.get(c, "") for c in cols])
    return table


@click.group()
def cli():
    pass


@cli.command("config", help="configure to access qcloud api")
@click.option("--secret_id", required=True, help="your secret id to access qcloud api")
@click.option(
    "--secret_key", required=True, help="your secret key to access qcloud api"
)
@click.option("--region", default="ap-guangzhou", help="your region name")
def config_create_cmd(secret_id, secret_key, region):
    cred_path_dir = os.path.dirname(cred_path)
    if not os.path.isdir(cred_path_dir):
        os.makedirs(cred_path_dir)
    obj = {}
    obj["secret_id"] = secret_id
    obj["secret_key"] = secret_key
    obj["region"] = region
    with open(cred_path, "w") as f:
        json.dump(obj, f, indent=2)
    click.echo(f"credential file saved to {cred_path}")


def init_client() -> Client:
    if not os.path.exists(cred_path):
        raise click.ClickException(
            'Please run "tikit config" to ' "configure the secret first"
        )
    with open(cred_path, "r") as f:
        cred = json.loads(f.read())
    return Client(cred["secret_id"], cred["secret_key"], region=cred["region"])


@cli.command("create", help="create tione resource")
@click.argument("resource", type=click.Choice(["service"]))
@click.option(
    "--init_file", is_flag=True, default=False, help="create a config file template"
)
@click.option(
    "-f", "--filename", default="", help="that contains the configuration to create"
)
def create_cmd(resource, init_file, filename):
    client = init_client()
    if resource == "service":
        try:
            if init_file is True:
                # 生成描述模版文件
                filename = filename if filename != "" else "service-config.yaml"
                dirname = os.path.dirname(__file__)
                with open(
                    os.path.join(dirname, "templates", "service_config.yaml"), "r"
                ) as infile:
                    config_template = infile.read()
                with open(filename, "w") as outfile:
                    outfile.write(config_template)
            else:
                msg = client._create_model_service_from_file(filename)
                click.echo(msg)
        except Exception as e:
            raise e
            # raise click.ClickException(str(e))


@cli.command("update", help="update tione resource")
@click.argument("resource", type=click.Choice(["service"]))
@click.option(
    "-f", "--filename", default="", help="that contains the configuration to create"
)
def update_cmd(resource, filename):
    client = init_client()
    if resource == "service":
        try:
            msg = client._modify_model_service_from_file(filename)
            click.echo(msg)
        except Exception as e:
            raise click.ClickException(str(e))


@cli.command("balance", help="balance the traffic")
@click.argument("name", metavar="<NAME>|<SERVICE_GROUP_ID>")
@click.argument("weights", nargs=2, metavar="<VERSION1>:<WEIGHT1> ...")
def balance_cmd(name, weights):
    if not weights:
        raise click.ClickException("weights shoud not be empty")
    m = {}
    import re

    for wei in weights:
        group = wei.split(":")
        if len(group) != 2:
            raise click.ClickException(f"{wei} is invalid, " f"it should be like v1:30")
        group[0] = re.sub("VERSION", "", group[0])
        group[0] = re.sub("version", "", group[0])
        m[group[0].lstrip("vV")] = int(group[1])
    try:
        client = init_client()
        service_group_id = client.get_service_group_id_by_name(name)
        mm = {}
        for v, w in m.items():
            mm[f"{service_group_id}-{v}"] = w
        resp = client.modify_service_group_weights(service_group_id, mm)
        click.echo(f"change weights succeeded: req_id = {resp.RequestId}")
    except Exception as e:
        raise click.ClickException(str(e))


@cli.command("stop", help="stop tione resource")
@click.argument("resource", type=click.Choice(["service"]))
@click.argument("name", metavar="<ID>|(<NAME>:<VERSION>)")
def stop_cmd(resource, name: str):
    client = init_client()
    if resource == "service":
        if ":" in name:
            group = name.split(":")
            if len(group) != 2:
                raise click.ClickException(
                    f"service name({name}) is invalid, " f"should be <NAME>:<VERSION>"
                )
            service_id = client.get_service_id_by_name(group[0], group[1])
        else:
            service_id = name
        try:
            msg = client.stop_model_service(service_id)
            click.echo(msg)
        except Exception as e:
            raise click.ClickException(str(e))


@cli.command("start", help="start tione resource")
@click.argument("resource", type=click.Choice(["service"]))
@click.argument("name", metavar="<ID>|(<NAME>:<VERSION>)")
def start_cmd(resource, name: str):
    client = init_client()
    if resource == "service":
        if ":" in name:
            group = name.split(":")
            if len(group) != 2:
                raise click.ClickException(
                    f"service name({name}) is invalid, " f"should be <NAME>:<VERSION>"
                )
            service_id = client.get_service_id_by_name(group[0], group[1])
        else:
            service_id = name
        try:
            msg = client.start_model_service(service_id)
            click.echo(msg)
        except Exception as e:
            raise click.ClickException(str(e))


@cli.command("delete", help="delete tione resource")
@click.argument("resource", type=click.Choice(["service"]))
@click.argument("name", metavar="<ID>|(<NAME>:<VERSION>)")
def delete_cmd(resource, name: str):
    client = init_client()
    if resource == "service":
        if ":" in name:
            group = name.split(":")
            if len(group) != 2:
                raise click.ClickException(
                    f"service name({name}) is invalid, " f"should be <NAME>:<VERSION>"
                )
            service_id = client.get_service_id_by_name(group[0], group[1])
        else:
            service_id = name
        try:
            msg = client.delete_model_service(service_id)
            click.echo(msg)
        except Exception as e:
            raise click.ClickException(str(e))


@cli.command("get", help="get info of tione resource")
@click.argument("resource", type=click.Choice(["service"]))
@click.argument("name", metavar="<ID>|(<NAME>:<VERSION>)", required=False)
def get_cmd(resource, name: str):
    client = init_client()
    if resource == "service":
        if not name:
            # 返回所有的服务列表
            svcs = client.get_model_service_summary()
            click.echo(pretty_print_dict_list(svcs))
            return
        if ":" in name:
            group = name.split(":")
            if len(group) != 2:
                raise click.ClickException(
                    f"service name({name}) is invalid, " f"should be <NAME>:<VERSION>"
                )
            service_id = client.get_service_id_by_name(group[0], group[1])
        else:
            service_id = name
        try:
            resp = client.describe_model_service(service_id)
            click.echo(resp)
        except Exception as e:
            raise click.ClickException(str(e))



#!/usr/bin/python
# -*- coding: utf-8 -*-
import logging
import threading

from requests.adapters import HTTPAdapter
from urllib3 import HTTPSConnectionPool, PoolManager, HTTPConnectionPool

logger = logging.getLogger("tencentcloud_sdk_common")


class HTTPSPreConnPool(HTTPSConnectionPool):
    _close_signal = {}

    def __init__(self, *args, **kwargs):
        super(HTTPSPreConnPool, self).__init__(*args, **kwargs)
        # clear the pool
        for _ in range(self.pool.maxsize):
            self.pool.get()
        self._conn_producer = threading.Thread(target=self._conn_producer_loop)
        self._conn_producer.setDaemon(True)
        self._conn_producer.start()

    def _conn_producer_loop(self):
        while True:
            conn = super(HTTPSPreConnPool, self)._new_conn()
            conn.connect()
            logger.debug("HTTPSPreConnPool: created a new conn")
            self.pool.put(conn)


class HTTPPreConnPool(HTTPConnectionPool):
    _close_signal = {}

    def __init__(self, *args, **kwargs):
        super(HTTPPreConnPool, self).__init__(*args, **kwargs)
        # clear the pool
        for _ in range(self.pool.maxsize):
            self.pool.get()
        self._conn_producer = threading.Thread(target=self._conn_producer_loop)
        self._conn_producer.setDaemon(True)
        self._conn_producer.start()

    def _conn_producer_loop(self):
        while True:
            conn = super(HTTPPreConnPool, self)._new_conn()
            conn.connect()
            logger.debug("HTTPSPreConnPool: created a new conn")
            self.pool.put(conn)


class PreConnPoolManager(PoolManager):
    def __init__(self, pool_size, *args, **kwargs):
        self._pool_size = pool_size
        super(PreConnPoolManager, self).__init__(*args, **kwargs)

    def _new_pool(self, scheme, host, port, request_context):
        if scheme == 'https':
            return HTTPSPreConnPool(host, port, maxsize=self._pool_size - 1)
        if scheme == 'http':
            return HTTPPreConnPool(host, port, maxsize=self._pool_size - 1)
        return super(PreConnPoolManager, self)._new_pool(scheme, host, port, request_context)


class PreConnAdapter(HTTPAdapter):
    def __init__(self, conn_pool_size, *args, **kwargs):
        self._conn_pool_size = conn_pool_size
        super(PreConnAdapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        self.poolmanager = PreConnPoolManager(self._conn_pool_size, *args, **kwargs)



# -*- coding: utf-8 -*-
__version__ = "1.9.1.250324"
from .default_client import *
from .display_optimize import *

# from .client import Client
# from .hdfs import *
# from .hive import *
# from .models import *



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import logging

import requests
import certifi

from tikit.tencentcloud.common.http.pre_conn import PreConnAdapter

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from tikit.tencentcloud.common.exception.tencent_cloud_sdk_exception import TencentCloudSDKException

logger = logging.getLogger("tencentcloud_sdk_common")


def _get_proxy_from_env(host, varname="HTTPS_PROXY"):
    no_proxy = os.environ.get("NO_PROXY") or os.environ.get("no_proxy")
    if no_proxy and host in no_proxy:
        return None
    return os.environ.get(varname.lower()) or os.environ.get(varname.upper())


class ProxyConnection(object):
    def __init__(self, host, timeout=60, proxy=None, certification=None, is_http=False, pre_conn_pool_size=0):
        self.request_host = host
        self.certification = certification
        if certification is None:
            self.certification = certifi.where()
        self.timeout = timeout
        self.proxy = None
        if is_http:
            proxy = proxy or _get_proxy_from_env(host, varname="HTTP_PROXY")
        else:
            proxy = proxy or _get_proxy_from_env(host, varname="HTTPS_PROXY")
        if proxy:
            self.proxy = {"http": proxy, "https": proxy}
        self.request_length = 0
        self._session = requests.Session()
        if pre_conn_pool_size > 0:
            adapter = PreConnAdapter(conn_pool_size=pre_conn_pool_size)
            self._session.mount("https://", adapter)
            self._session.mount("http://", adapter)

    def request(self, method, url, body=None, headers=None):
        headers.setdefault("Host", self.request_host)
        return self._session.request(method=method,
                                     url=url,
                                     data=body,
                                     headers=headers,
                                     proxies=self.proxy,
                                     verify=self.certification,
                                     timeout=self.timeout,
                                     stream=True)


class ApiRequest(object):
    def __init__(self, host, req_timeout=60, debug=False, proxy=None, is_http=False, certification=None,
                 pre_conn_pool_size=0):
        self.conn = ProxyConnection(host, timeout=req_timeout, proxy=proxy, certification=certification,
                                    is_http=is_http, pre_conn_pool_size=pre_conn_pool_size)
        self.is_http = is_http
        self.host = host
        self.req_timeout = req_timeout
        self.keep_alive = False
        self.debug = debug
        self.request_size = 0
        self.response_size = 0

    def _handle_host(self, host):
        url = urlparse(host)
        if not url.hostname:
            if self.is_http:
                return "http://" + host
            else:
                return "https://" + host
        return host

    def set_req_timeout(self, req_timeout):
        self.req_timeout = req_timeout

    def is_keep_alive(self):
        return self.keep_alive

    def set_keep_alive(self, flag=True):
        self.keep_alive = flag

    def set_debug(self, debug):
        self.debug = debug

    def _request(self, req_inter):
        url = self._handle_host(req_inter.host)
        if self.keep_alive:
            req_inter.header["Connection"] = "Keep-Alive"
        logger.debug("SendRequest: %s" % req_inter)
        if req_inter.method == 'GET':
            req_inter_url = '%s?%s' % (url, req_inter.data)
            return self.conn.request(req_inter.method, req_inter_url,
                                     None, req_inter.header)
        elif req_inter.method == 'POST':
            return self.conn.request(req_inter.method, url,
                                     req_inter.data, req_inter.header)
        else:
            raise TencentCloudSDKException(
                "ClientParamsError", 'Method only support (GET, POST)')

    def send_request(self, req_inter):
        try:
            http_resp = self._request(req_inter)
            self.request_size = self.conn.request_length
            return http_resp
        except Exception as e:
            raise TencentCloudSDKException("ClientNetworkError", str(e))


class RequestInternal(object):
    def __init__(self, host="", method="", uri="", header=None, data=""):
        if header is None:
            header = {}
        self.host = host
        self.method = method
        self.uri = uri
        self.header = header
        self.data = data

    def __str__(self):
        headers = "\n".join("%s: %s" % (k, v) for k, v in self.header.items())
        return ("Host: %s\nMethod: %s\nUri: %s\nHeader: %s\nData: %s\n"
                % (self.host, self.method, self.uri, headers, self.data))


class ResponsePrettyFormatter(object):
    def __init__(self, resp, format_body=True, delimiter="\n"):
        self._resp = resp
        self._format_body = format_body
        self._delimiter = delimiter

    def __str__(self):
        lines = ['%s %d %s' % (self.str_ver(self._resp.raw.version), self._resp.status_code, self._resp.reason)]
        for k, v in self._resp.headers.items():
            lines.append('%s: %s' % (k, v))
        if self._format_body:
            lines.append('')
            lines.append(self._resp.text)
        return self._delimiter.join(lines)

    @staticmethod
    def str_ver(ver):
        if ver == 10:
            return "HTTP/1.0"
        elif ver == 11:
            return "HTTP/1.1"
        elif ver == 20:
            return "HTTP/2.0"
        else:
            return str(ver)



import types
import os
import re
import shutil

from tikit.tencentcloud.tione.v20211111 import models

# 日志查询条数的最大值
LIMIT = 1000


class KitLog(object):
    
    def __init__(self, tione_client):
        self._tione_client = tione_client

    def save(self, path, compress=None, service=None, service_id=None, pod_name=None, start_time=None, end_time=None, order=None):
        # 安全处理 service_id 为空的情况
        base_dir = str(service_id) if service_id is not None else "unknown_service"

        # 创建唯一目标目录
        target_dir = os.path.join(path, base_dir)
        suffix = 1
        while os.path.exists(target_dir):
            target_dir = os.path.join(path, f"{base_dir}_{suffix}")
            suffix += 1
        os.makedirs(target_dir)

        # 准备文件句柄字典和文件名映射
        file_handles = {}
        sanitized_names = {}

        # 清理文件名非法字符的函数
        def sanitize_filename(name):
            return re.sub(r'[\\/*?:"<>|]', "_", name)

        try:
            # 初始化分页查询
            result = self.download(service, service_id, pod_name, start_time, end_time, LIMIT, order)

            # 处理时间格式（用于文件名）
            start_str = format_time_for_filename(start_time) if start_time else "None"
            end_str = format_time_for_filename(end_time) if end_time else "None"

            # 分页获取所有日志
            while result is not None and result.Content is not None:
                for log in result.Content:
                    # 获取Pod名称（处理空值）
                    pod = log.PodName.strip() if log.PodName and log.PodName.strip() else "unknown_pod"

                    # 清理特殊字符并创建唯一文件名
                    safe_pod = sanitize_filename(pod)
                    if safe_pod not in sanitized_names:
                        # 生成实际文件名（带时间范围）
                        filename = f"{safe_pod}-{start_str}-{end_str}.log"
                        sanitized_names[safe_pod] = filename
                        file_path = os.path.join(target_dir, filename)

                        # 以追加模式打开文件
                        file_handles[safe_pod] = open(file_path, "a", encoding="utf-8")

                    # 写入日志内容（确保每行一个日志）
                    if log.Message:
                        # 移除行尾换行符后添加统一换行
                        message = log.Timestamp + " " + log.Message.rstrip("\n") + "\n"
                        file_handles[safe_pod].write(message)

                # 获取下一页
                result = result.next()

        finally:
            # 确保所有文件句柄正确关闭
            for handle in file_handles.values():
                handle.close()
            # 新增的压缩功能
            if compress:
                # 创建压缩文件的路径
                compress_dir = os.path.dirname(target_dir)  # 上级目录
                compress_name = os.path.basename(target_dir)  # 目录名作为压缩文件名
                compress_path = os.path.join(compress_dir, compress_name)

                # 创建压缩文件
                shutil.make_archive(compress_path, 'zip', root_dir=compress_dir, base_dir=compress_name)

                # 移除原始文件夹
                shutil.rmtree(target_dir)
                print(f"已压缩日志文件到: {compress_path}.zip")

    def download(
        self,
        service=None,
        service_id=None,
        pod_name=None,
        start_time=None,
        end_time=None,
        limit=None,
        order=None,
        context=None,
    ):
        """下载日志文件到本地

        :param service: 下载哪种类型的日志。 TRAIN：训练任务； NOTEBOOK：Notebook服务； INFER：推理服务；
        :type service: str
        :param service_id: 服务ID
        :type service: str
        :param pod_name: 查询哪个Pod的日志，支持支持结尾通配符。查看某个训练任务的全部pod的日志可以填： "<task_id>*"，如：train-51cd6bf7ec1000*
        :type pod_name: str
        :param start_time: 日志查询开始时间。RFC3339格式的时间字符串，比如2021-12-16T13:20:24+08:00，默认值为当前时间的前一个小时
        :type start_time: str
        :param end_time: 日志查询结束时间。RFC3339格式的时间字符串，比如2021-12-16T13:20:24+08:00，默认值为当前时间
        :type end_time: str
        :param limit: 日志查询条数，默认值100，最大值100
        :type limit: int
        :param order: 排序方向。(ASC | DESC) 默认值为DESC
        :type order: str
        :param context: 分页的游标
        :type context: str
        :rtype: :class:`tikit.tencentcloud.tione.v20211111.models.DescribeLogsResponse`

            返回的对象如果非空，就会有 next() 方法，能不断地获取下一页的日志（如果有多页的话），如下：
            now_time = datetime.datetime.now(datetime.timezone.utc)
            now_time_str = now_time.isoformat()
            result = client.download_logs("train-51cd6bf7ec1000-37c5p5nlr01s-launcher",
                                                "2021-12-10T09:32:03.823509+00:00",
                                                now_time_str,
                                                limit=30)
            print(result)
            print(result.next())
            print(result.next())
            print(result.next())

        """
        if order is None:
            order = "ASC"

        req = models.DescribeLogsRequest()
        req.Service = service
        req.ServiceId = service_id
        req.PodName = pod_name
        req.StartTime = start_time
        req.EndTime = end_time
        req.Limit = limit
        req.Order = order
        req.Context = context
        result = self._tione_client.DescribeLogs(req)

        def get_next_data(xx):
            if result.Context != "":
                next_result = self.download(
                    service,
                    service_id,
                    pod_name,
                    start_time,
                    end_time,
                    limit,
                    order,
                    result.Context,
                )
                result.Context = next_result.Context
                result.Content = next_result.Content
                result.RequestId = next_result.RequestId
                return result
            else:
                print("Download finished")
                return None

        result.next = types.MethodType(get_next_data, result)
        return result


def format_time_for_filename(time_str):
    if not time_str:
        return "None"
    # 移除时区信息和特殊字符，转换为YYYYMMDDHHMMSS格式
    return time_str[:19].replace("-", "").replace(":", "").replace("T", "")


# Copyright (c) 2018 Tencent Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import warnings
from tikit.tencentcloud.common.exception.tencent_cloud_sdk_exception import TencentCloudSDKException
from tikit.tencentcloud.common.profile.http_profile import HttpProfile


class ClientProfile(object):
    unsignedPayload = False

    def __init__(self, signMethod=None, httpProfile=None, language="zh-CN",
                 disable_region_breaker=True, region_breaker_profile=None, request_client=None):
        """SDK profile.

        :param signMethod: The signature method, valid choice: HmacSHA1, HmacSHA256, TC3-HMAC-SHA256
        :type signMethod: str
        :param httpProfile: The http profile
        :type httpProfile: :class:`HttpProfile`
        :param language: Valid choice: en-US, zh-CN.
        :type language: str
        :param disable_region_breaker: Switch of region breaker.
        :type disable_region_breaker: bool
        :param region_breaker_profile: The region breaker profile.
        :type region_breaker_profile: :class:`RegionBreakerProfile`
        :param request_client: Custom request client.
        :type request_client: str
        """
        self.httpProfile = HttpProfile() if httpProfile is None else httpProfile
        self.signMethod = "TC3-HMAC-SHA256" if signMethod is None else signMethod
        valid_language = ["zh-CN", "en-US"]
        if language not in valid_language:
            raise TencentCloudSDKException("ClientError", "Language invalid, choices: %s" % valid_language)
        self.language = language
        self.disable_region_breaker = disable_region_breaker
        self.region_breaker_profile = region_breaker_profile
        if not self.disable_region_breaker and self.region_breaker_profile is None:
            self.region_breaker_profile = RegionBreakerProfile()
        self.request_client = None
        if isinstance(request_client, str) and re.match("^[0-9a-zA-Z-_,;.]+$", request_client):
            if len(request_client) > 128:
                warnings.warn("the length of RequestClient should be with in 128 characters, it will be truncated")
            self.request_client = request_client[:128]
        elif request_client is not None:
            warnings.warn("RequestClient not match the regexp: ^[0-9a-zA-Z-_,;.]+$, ignored")


class RegionBreakerProfile(object):

    def __init__(self, backup_endpoint="ap-guangzhou.tencentcloudapi.com",
                 max_fail_num=5,
                 max_fail_percent=0.75,
                 window_interval=60*5,
                 timeout=60,
                 max_requests=5):
        """RegionBreaker profile.

        :param backup_endpoint: Backup region of endpoint, default ap-guangzhou.tencentcloudapi.com.
        :type backup_endpoint: str
        :param max_fail_num: Max failure to trigger breaker, default 5.
        :type max_fail_num: int
        :param max_fail_percent: Max failure percent to trigger breaker, default 0.75.
        :type max_fail_percent: float
        :param window_interval: Decides when to reset counter if the state is StateClosed, default 5 minutes.
        :type window_interval: int
        :param timeout: Decides when to turn StateOpen to StateHalfOpen, default 1 minutes.
        :type timeout: int
        :param max_requests: Decides when to turn StateHalfOpen to StateClosed, default 5 successful requests.
        :type timeout: int
        """
        self.backup_endpoint = backup_endpoint
        if not self.check_endpoint():
            raise TencentCloudSDKException("ClientError", "the format of `backup_endpoint` must be tencentcloudapi.com "
                                                          "or ${region}.tencentcloudapi.com")
        self.max_fail_num = max_fail_num
        self.max_fail_percent = max_fail_percent
        if self.max_fail_percent < 0 or self.max_fail_percent > 1:
            raise TencentCloudSDKException("ClientError", "max fail percent must be set between 0 and 1")
        self.window_interval = window_interval
        self.timeout = timeout
        self.max_requests = max_requests

    def check_endpoint(self):
        endpoint_split = self.backup_endpoint.split(".")
        if len(endpoint_split) != 3 and len(endpoint_split) != 2:
            return False
        if endpoint_split[-2] != "tencentcloudapi" or endpoint_split[-1] != "com":
            return False
        if len(endpoint_split) == 3:
            region = endpoint_split[0]
            if len(region.split("-")) != 2:
                return False
        return True




# -*- coding: utf-8 -*-
#
# Copyright 1999-2017 Tencent Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import threading


STATE_CLOSED = 0
STATE_HALF_OPEN = 1
STATE_OPEN = 2


class Counter(object):

    def __init__(self):
        self.failures = 0
        self.total = 0
        self.consecutive_successes = 0
        self.consecutive_failures = 0

    def on_success(self):
        self.total += 1
        self.consecutive_successes += 1
        self.consecutive_failures = 0

    def on_failure(self):
        self.total += 1
        self.failures += 1
        self.consecutive_failures += 1
        self.consecutive_successes = 0

    def clear(self):
        self.failures = 0
        self.total = 0
        self.consecutive_successes = 0
        self.consecutive_failures = 0

    def get_failure_rate(self):
        if self.total == 0:
            return 0.0
        return float(self.failures) / self.total


class CircuitBreaker(object):

    def __init__(self, breaker_setting):
        self.breaker_setting = breaker_setting
        self.lock = threading.Lock()
        self.counter = Counter()
        self.state = STATE_CLOSED
        self.expiry = time.time() + breaker_setting.window_interval
        self.generation = 0

    def ready_to_open(self):
        return (self.counter.failures >= self.breaker_setting.max_fail_num and
                self.counter.get_failure_rate() >= self.breaker_setting.max_fail_percent) or \
               self.counter.consecutive_failures >= 5

    def current_state(self, now):
        if self.state == STATE_CLOSED:
            if self.expiry <= now:
                self.to_new_generation(now)
        elif self.state == STATE_OPEN:
            if self.expiry <= now:
                self.switch_state(STATE_HALF_OPEN, now)
        return self.state, self.generation

    def switch_state(self, new_state, now):
        if self.state == new_state:
            return
        self.state = new_state
        self.to_new_generation(now)

    def to_new_generation(self, now):
        self.generation = (self.generation + 1) % 10
        self.counter.clear()
        if self.state == STATE_CLOSED:
            self.expiry = now + self.breaker_setting.window_interval
        elif self.state == STATE_OPEN:
            self.expiry = now + self.breaker_setting.timeout
        else:  # STATE_HALF_OPEN
            self.expiry = time.time()

    # whether to use the backup region
    def before_requests(self):
        self.lock.acquire()
        now = time.time()
        state, generation = self.current_state(now)
        self.lock.release()
        if state == STATE_OPEN:
            return generation, True
        return generation, False

    def after_requests(self, before, success):
        self.lock.acquire()
        now = time.time()
        state, generation = self.current_state(now)
        self.lock.release()
        # the breaker has entered the next generation, the current results abandon.
        if generation != before:
            return
        if success:
            self.on_success(state, now)
        else:
            self.on_failure(state, now)

    def on_success(self, state, now):
        if state == STATE_CLOSED:
            self.counter.on_success()
        elif state == STATE_HALF_OPEN:
            self.counter.on_success()
            if self.counter.total - self.counter.failures >= self.breaker_setting.max_requests:
                self.switch_state(STATE_CLOSED, now)

    def on_failure(self, state, now):
        if state == STATE_CLOSED:
            self.counter.on_failure()
            if self.ready_to_open():
                self.switch_state(STATE_OPEN, now)
        elif state == STATE_HALF_OPEN:
            self.counter.on_failure()
            self.switch_state(STATE_OPEN, now)
